from __future__ import annotations

from collections import deque
from io import BytesIO
import os
import re
import textwrap
import threading
import time as ttime
import uuid
from enum import Enum
from pathlib import Path

import numpy as np
import requests
from caproto import ChannelType
from caproto.ioc_examples.mini_beamline import no_reentry
from caproto.server import PVGroup, pvproperty, run, template_arg_parser
from ophyd import Component as Cpt
from ophyd import Device, EpicsSignal, EpicsSignalRO, Kind, Signal
from ophyd.status import SubscriptionStatus
from event_model import compose_resource

from .utils import now, save_hdf5_nd, save_image

from PIL import Image


class AcqStatuses(Enum):
    """Enum class for acquisition statuses."""

    IDLE = "idle"
    ACQUIRING = "acquiring"

class DirExistsStatuses(Enum):
    DOES_NOT_EXIST = "does not exist"
    PERMISSION_ERROR = "insufficient permissions"
    EXISTS = "exists"


class UIDOptions(Enum):
    NONE = "none"
    SHORT = "short"
    FULL = "full"

class OnOffStates(Enum):
    DISABLE = "disable"
    ENABLE = "enable"

class CaprotoSaveIOC(PVGroup):
    """Generic Caproto Save IOC"""

    write_dir = pvproperty(
        value="/tmp",
        doc="The directory to write data to. It support datetime formatting, e.g. '/tmp/det/%Y/%m/%d/'",
        string_encoding="utf-8",
        dtype=ChannelType.CHAR,
        max_length=255,
    )
    file_name = pvproperty(
        value="test.h5",
        doc="The file name of the file to write to. It support <str>.format() based formatting, e.g. 'scan_{num:06d}.h5'",
        string_encoding="utf-8",
        dtype=ChannelType.CHAR,
        max_length=255,
    )
    full_file_path = pvproperty(
        value="",
        doc="Full path to the data file",
        dtype=str,
        read_only=True,
        max_length=255,
    )

    directory_exists = pvproperty(
        value=DirExistsStatuses.DOES_NOT_EXIST.value,
        doc="Record specifying whether or not the target directory exists or not",
        dtype=ChannelType.ENUM,
        read_only=True,
        enum_strings=[x.value for x in DirExistsStatuses]
    )

    uid_type = pvproperty(
        value=UIDOptions.NONE.value,
        doc="UUID to include automatically in file name",
        dtype=ChannelType.ENUM,
        enum_strings=[x.value for x in UIDOptions]
    )

    use_frame_num = pvproperty(
        value = OnOffStates.DISABLE.value,
        doc = "Enable auto-incrementing frame counter suffix for filenames",
        dtype=ChannelType.ENUM,
        enum_strings=[x.value for x in OnOffStates]
    )

    # TODO: check non-negative value in @frame_num.putter.
    frame_num = pvproperty(value=0, doc="Frame counter", dtype=int)

    acquire = pvproperty(
        value=AcqStatuses.IDLE.value,
        enum_strings=[x.value for x in AcqStatuses],
        dtype=ChannelType.ENUM,
        doc="Acquire signal to save a dataset.",
    )

    def __init__(self, *args, update_rate=10.0, **kwargs):
        super().__init__(*args, **kwargs)

        self._update_rate = update_rate
        self._update_period = 1.0 / update_rate

        self._request_queue = None
        self._response_queue = None

        self._sanitizer = re.compile(pattern=r"[\":<>|\*\?\s]")

    queue = pvproperty(value=0, doc="A PV to facilitate threading-based queue")

    @queue.startup
    async def queue(self, instance, async_lib):
        """The startup behavior of the count property to set up threading queues."""
        # pylint: disable=unused-argument
        self._request_queue = async_lib.ThreadsafeQueue(maxsize=1)
        self._response_queue = async_lib.ThreadsafeQueue(maxsize=1)

        # Start a separate thread that consumes requests and sends responses.
        thread = threading.Thread(
            target=self.saver,
            daemon=True,
            kwargs={
                "request_queue": self._request_queue,
                "response_queue": self._response_queue,
            },
        )
        thread.start()

    async def _update_full_file_path(self, write_dir=None, file_name=None, use_frame_num=None, uid_type=None):
        
        if use_frame_num is None:
            use_num = self.use_frame_num.value
        else:
            use_num = use_frame_num
        
        frame_num_str = ""
        if use_num == OnOffStates.ENABLE.value:
            frame_num = self.frame_num.value
            frame_num_str = f"_{frame_num:06}"

        if uid_type is None:
            uid_to_use = self.uid_type.value
        else:
            uid_to_use = uid_type
        uid_str = ""
        if uid_to_use == UIDOptions.SHORT.value:
            uid_str = f"_{str(uuid.uuid4())[:8]}"
        elif uid_to_use == UIDOptions.FULL.value:
            uid_str = f"_{str(uuid.uuid4())}"


        if write_dir is None:
            local_write_dir = Path(self.write_dir.value)
        else:
            local_write_dir = Path(write_dir)

        if file_name is None:
            full_file_name = self.file_name.value
        else:
            full_file_name = file_name

        try:
            filename_and_ext = os.path.splitext(full_file_name)
            base_filename = filename_and_ext[0]
            extension = filename_and_ext[1]
        except IndexError:
            # Case that we didn't get an extension
            base_filename = full_file_name
            extension = ""

        full_file_path = local_write_dir / f"{base_filename}{frame_num_str}{uid_str}{extension}"

        full_file_path = self._sanitizer.sub("_", str(full_file_path))

        print(f"{now()}: {full_file_path = }")

        await self.full_file_path.write(full_file_path)


    async def _use_frame_num_callback(self, instance, value):
        await self._update_full_file_path(use_frame_num=value)
        return value

    async def _uid_type_callback(self, instance, value):
        await self._update_full_file_path(uid_type=value)
        return value

    async def _file_name_callback(self, instance, value):
        await self._update_full_file_path(file_name=value)
        return value

    async def _write_dir_callback(self, instance, value):
        """The stage method to perform preparation of a dataset to save the data."""

        local_write_dir = Path(value)

        if os.path.exists(local_write_dir):
            if os.access(local_write_dir, os.W_OK):
                await self.directory_exists.write(DirExistsStatuses.EXISTS.value)
                await self._update_full_file_path(write_dir=value)
            else:
                await self.directory_exists.write(DirExistsStatuses.PERMISSION_ERROR.value)
        else:
            await self.directory_exists.write(DirExistsStatuses.DOES_NOT_EXIST.value)


        if self.directory_exists.value == DirExistsStatuses.EXISTS.value:
            return value
        else:
            print(f"Directory access error for directory {value}! - {self.directory_exists.value}")
            return ""

    @write_dir.putter
    async def write_dir(self, *args, **kwargs):
        """The write_dir callback method."""
        return await self._write_dir_callback(*args, **kwargs)

    @file_name.putter
    async def file_name(self, *args, **kwargs):
        """The file name callback method."""
        return await self._file_name_callback(*args, **kwargs)
    

    @uid_type.putter
    async def uid_type(self, *args, **kwargs):
        """The file name callback method."""
        return await self._uid_type_callback(*args, **kwargs)
    

    @use_frame_num.putter
    async def use_frame_num(self, *args, **kwargs):
        """The file name callback method."""
        return await self._use_frame_num_callback(*args, **kwargs)

    async def _get_current_dataset(self, frame):
        """The method to return a desired dataset.

        See https://scikit-image.org/docs/stable/auto_examples/data/plot_3d.html
        for details about the dataset returned by the base class' method.
        """
        return np.random.random((480, 640))


    @acquire.putter
    @no_reentry
    async def acquire(self, instance, value):
        """The acquire method to perform an individual acquisition of a data point."""
        if (
            value != AcqStatuses.ACQUIRING.value
            # or self.stage.value not in [True, StageStates.STAGED.value]
        ):
            return False

        if (
            instance.value in [True, AcqStatuses.ACQUIRING.value]
            and value == AcqStatuses.ACQUIRING.value
        ):
            print(
                f"The device is already acquiring. Please wait until the '{AcqStatuses.IDLE.value}' status."
            )
            return True

        if (self.directory_exists.value != DirExistsStatuses.EXISTS.value):
            print("Target write directory does not exist or cannot be written to!")
            return False

        await self.acquire.write(AcqStatuses.ACQUIRING.value)

        # Delegate saving the resulting data to a blocking callback in a thread.
        payload = {
            "filename": self.full_file_path.value,
            "data": await self._get_current_dataset(frame=self.frame_num.value),
            "uid": str(uuid.uuid4()),
            "timestamp": ttime.time(),
            "frame_number": self.frame_num.value,
        }

        await self._request_queue.async_put(payload)
        response = await self._response_queue.async_get()

        if response["success"]:
            # Increment the counter only on a successful saving of the file.
            await self.frame_num.write(self.frame_num.value + 1)

        # await self.acquire.write(AcqStatuses.IDLE.value)

        return False


    async def on_startup(self, async_lib):
        for key in self.pvdb:
            print(key)

        await self._write_dir_callback(None, "/tmp")

    @staticmethod
    def saver(request_queue, response_queue):
        """The saver callback for threading-based queueing."""
        while True:
            received = request_queue.get()
            filename = received["filename"]
            data = received["data"]
            frame_number = received["frame_number"]
            try:
                save_hdf5_nd(fname=filename, data=data, mode="x", group_path="enc1")
                print(
                    f"{now()}: saved {frame_number=} {data.shape} data into:\n  {filename}"
                )

                success = True
                error_message = ""
            except Exception as exc:  # pylint: disable=broad-exception-caught
                success = False
                error_message = exc
                print(
                    f"Cannot save file {filename!r} due to the following exception:\n{exc}"
                )

            response = {"success": success, "error_message": error_message}
            response_queue.put(response)

    @staticmethod
    def check_args(parser, split_args):
        """Helper function to process caproto CLI args."""
        parsed_args = parser.parse_args()
        prefix = parsed_args.prefix
        if not prefix:
            parser.error("The 'prefix' argument must be specified.")

        ioc_opts, run_opts = split_args(parsed_args)
        return ioc_opts, run_opts

class AxisWebcamCaprotoSaver(CaprotoSaveIOC):
    """"""

    def __init__(self, *args, camera_host=None, **kwargs):
        self._camera_host = camera_host
        print(f"{camera_host = }")
        super().__init__(*args, **kwargs)

    @staticmethod
    def check_args(parser, split_args):
        """Helper function to process caproto CLI args."""
        parsed_args = parser.parse_args()
        prefix = parsed_args.prefix
        camera_host = parsed_args.camera_host
        if not prefix:
            parser.error("The 'prefix' argument must be specified.")
        if not camera_host:
            parser.error("The 'camera_host' argument must be specified.")

        ioc_opts, run_opts = split_args(parsed_args)

        ioc_opts['camera_host'] = parsed_args.camera_host
        return ioc_opts, run_opts


    async def _get_current_dataset(self, *args, **kwargs):  # pylint: disable=unused-argument
        url = f"http://{self._camera_host}/axis-cgi/jpg/image.cgi"
        resp = requests.get(url, timeout=10)
        img = Image.open(BytesIO(resp.content))

        dataset = np.asarray(img).sum(axis=-1)
        print(f"{now()}: {dataset.shape}")

        return dataset

    @staticmethod
    def saver(request_queue, response_queue):
        """The saver callback for threading-based queueing."""
        while True:
            received = request_queue.get()
            filename = received["filename"]
            data = received["data"]
            # 'frame_number' is not used for this exporter.
            try:
                save_hdf5_nd(fname=filename, data=data, dtype="|u1", mode="a")
                #TODO: Change all of these prints to use the caproto logger instead
                print(f"{now()}: saved data into: {filename}")

                success = True
                error_message = ""
            except Exception as exc:  # pylint: disable=broad-exception-caught
                success = False
                error_message = exc
                print(
                    f"Cannot save file {filename!r} due to the following exception:\n{exc}"
                )

            response = {"success": success, "error_message": error_message}
            response_queue.put(response)


class ExternalFileReference(Signal):
    """
    A pure software Signal that describe()s an image in an external file.
    """

    def describe(self):
        resource_document_data = super().describe()
        resource_document_data[self.name].update(
            dict(
                external="FILESTORE:",
                dtype="array",
            )
        )
        return resource_document_data


class CaprotoSaverDevice(Device):
    """An ophyd Device which works with the base caproto extension IOC."""

    write_dir = Cpt(EpicsSignal, "write_dir", string=True)
    file_name = Cpt(EpicsSignal, "file_name", string=True)
    full_file_path = Cpt(EpicsSignalRO, "full_file_path", string=True)
    frame_num = Cpt(EpicsSignal, "frame_num")
    acquire = Cpt(EpicsSignal, "acquire", string=True)
    directory_exists = Cpt(EpicsSignalRO, "directory_exists", string=True)
    use_frame_num = Cpt(EpicsSignal, "use_frame_num", string=True)
    uid_type = Cpt(EpicsSignal, "uid_type", string=True)

    data = Cpt(ExternalFileReference, kind=Kind.normal)

    def __init__(self, *args, md=None, extension='h5', handler_spec="AD_HDF5", root_dir=None, **kwargs):
        super().__init__(*args, **kwargs)
        if root_dir is None:
            msg = "The 'root_dir' kwarg cannot be None"
            raise RuntimeError(msg)
        self._root_dir = root_dir
        self._resource_document, self._datum_factory = None, None
        self._asset_docs_cache = deque()
        self._md = md or {}
        self._handler_spec = handler_spec

        self.stage_sigs["uid_type"] = "full"
        self.stage_sigs["file_name"] = f"{self.name}.{extension}"

    def _update_paths(self):
        self._root_dir = self.root_path_str

    @property
    def root_path_str(self):
        beamline = os.getenv("ENDSTATION_ACRONYM", os.getenv("BEAMLINE_ACRONYM", "TST")).lower()
        # These three beamlines have a -new suffix in their 
        if beamline in ["xpd", "fxi", "qas"]:
            beamline = f"{beamline}-new"
        root_path = f"/nsls2/data/{beamline}/proposals/{self._md.get('cycle', '')}/{self._md.get('data_session', '')}/assets/{self.name}"
        return root_path
    
    @property
    def shape(self):
        """Property that contains the shape of the data"""

        return (1080, 1920)
    
    @property
    def dtype_numpy(self):
        """dtype_str for use in the descriptor"""
        return "<f4"

    def collect_asset_docs(self):
        """The method to collect resource/datum documents."""
        items = list(self._asset_docs_cache)
        self._asset_docs_cache.clear()
        yield from items

    def stage(self):
        self._update_paths()
        self.write_dir.put(self._root_dir)

        assert self.directory_exists.get() == "exists"

        super().stage()

        # Clear asset docs cache which may have some documents from the previous failed run.
        self._asset_docs_cache.clear()

        self._resource_document, self._datum_factory, _ = compose_resource(
            start={"uid": "needed for compose_resource() but will be discarded"},
            spec=self._handler_spec,
            root="/",
            resource_path=self.full_file_path.get(),
            resource_kwargs={},
        )

        # now discard the start uid, a real one will be added later
        self._resource_document.pop("run_start")
        self._asset_docs_cache.append(("resource", self._resource_document))

        # Update caproto IOC parameters:


    def describe(self):
        res = super().describe()
        res[self.data.name].update(
            {"shape": self.shape, "dtype_str": self.dtype_numpy}
        )
        return res


    def trigger(self):

        def done_callback(value, old_value, **kwargs):
            """The callback function used by ophyd's SubscriptionStatus."""
            # print(f"{old_value = } -> {value = }")
            if old_value == "acquiring" and value == "idle":
                return True
            return False

        status = SubscriptionStatus(self.acquire, run=False, callback=done_callback)

        # Reuse the counter from the caproto IOC
        self.acquire.put(1)

        datum_document = self._datum_factory(datum_kwargs={})
        self._asset_docs_cache.append(("datum", datum_document))

        self.data.put(datum_document["datum_id"])

        return status

    def unstage(self):
        self._resource_document = None
        self._datum_factory = None
        super().unstage()


class TwoDimCaprotoCam(CaprotoSaverDevice):


    def __init__(self, *args, shape=(1080, 1920), dtype_numpy="|u1", **kwargs):
        super().__init__(*args, **kwargs)
        self._shape = shape
        self._dtype_numpy = dtype_numpy

    @property
    def shape(self):
        return self._shape

    @property
    def dtype_numpy(self):
        return self._dtype_numpy


def start_caproto_ioc(cls, parser, split_args):

    ioc_options, run_options = cls.check_args(parser, split_args)
    ioc = cls(**ioc_options)
    run(ioc.pvdb, startup_hook=ioc.on_startup, **run_options)

def start_axis_ioc():
    parser, split_args = template_arg_parser(
        default_prefix="", desc=textwrap.dedent(AxisWebcamCaprotoSaver.__doc__)
    )
    parser.add_argument("-c", "--camera-host", help="URL of the axis camera stream", required=True, type=str)
    start_caproto_ioc(AxisWebcamCaprotoSaver, parser, split_args)


if __name__ == "__main__":
    start_axis_ioc()


from __future__ import annotations

import datetime
from pathlib import Path

import h5py
import numpy as np


def now(as_object=False):
    """A helper function to return ISO 8601 formatted datetime string."""
    _now = datetime.datetime.now()
    if as_object:
        return _now
    return _now.isoformat()


def save_image(fname, data, file_format="jpeg", mode="x"):  # pylint: disable=unused-argument
    """The function to export the image data (e.g., to a JPEG file."""
    data.save(fname, file_format=file_format)


def save_hdf5_zebra(
    fname,
    data,
    dtype="float32",
    mode="x",
):
    """The function to export the 1-D data to an HDF5 file.

    Check https://docs.h5py.org/en/stable/high/file.html#opening-creating-files for modes:

        r           Readonly, file must exist (default)
        r+          Read/write, file must exist
        w           Create file, truncate if exists
        w- or x     Create file, fail if exists
        a           Read/write if exists, create otherwise
    """
    with h5py.File(fname, mode, libver="latest") as h5file_desc:
        for pvname, value in data.items():
            dataset = h5file_desc.create_dataset(
                pvname,
                data=value,
                dtype=dtype,
            )
            dataset.flush()


def save_hdf5_nd(
    fname,
    data,
    group_name="/entry",
    group_path="data/data",
    dtype="float32",
    mode="x",
):
    """The function to export the N-D data to an HDF5 file (N>1).

    Check https://docs.h5py.org/en/stable/high/file.html#opening-creating-files for modes:

        r           Readonly, file must exist (default)
        r+          Read/write, file must exist
        w           Create file, truncate if exists
        w- or x     Create file, fail if exists
        a           Read/write if exists, create otherwise
    """
    update_existing = Path(fname).is_file()
    with h5py.File(fname, mode, libver="latest") as h5file_desc:
        frame_shape = data.shape
        if not update_existing:
            group = h5file_desc.create_group(group_name)
            dataset = group.create_dataset(
                group_path,
                data=np.full(fill_value=np.nan, shape=(1, *frame_shape)),
                maxshape=(None, *frame_shape),
                chunks=(1, *frame_shape),
                dtype=dtype,
            )
            frame_num = 0
        else:
            dataset = h5file_desc[f"{group_name}/{group_path}"]
            frame_num = dataset.shape[0]

        # https://docs.h5py.org/en/stable/swmr.html
        h5file_desc.swmr_mode = True

        dataset.resize((frame_num + 1, *frame_shape))
        dataset[frame_num, ...] = data
        dataset.flush()


from .providers import (
    YMDGranularity,
    ProposalNumScanNumPathProvider,
    ProposalNumYMDPathProvider,
    ShortUUIDFilenameProvider,
    AcqModeFilenameProvider,
    DeviceNameFilenameProvider,
    NSLS2PathProvider,
)

__all__ = [
    "YMDGranularity"
    "ProposalNumScanNumPathProvider",
    "ProposalNumYMDPathProvider",
    "ShortUUIDFilenameProvider",
    "AcqModeFilenameProvider",
    "DeviceNameFilenameProvider",
    "NSLS2PathProvider",
]



import argparse
import json
import os
import re
import warnings
from datetime import datetime
from getpass import getpass
from typing import Any, Dict

import httpx
import redis
import yaml
from ldap3 import NTLM, Connection, Server
from ldap3.core.exceptions import LDAPInvalidCredentialsResult, LDAPSocketOpenError
from redis_json_dict import RedisJSONDict

data_session_re = re.compile(r"^pass-(?P<proposal_number>\d+)$")

nslsii_api_client = httpx.Client(base_url="https://api.nsls2.bnl.gov")


def get_current_cycle() -> str:
    cycle_response = nslsii_api_client.get(
        f"/v1/facility/nsls2/cycles/current"
    ).raise_for_status()
    return cycle_response.json()["cycle"]


def is_commissioning_proposal(proposal_number, beamline) -> bool:
    """True if proposal_number is registered as a commissioning proposal; else False."""
    commissioning_proposals_response = nslsii_api_client.get(
        f"/v1/proposals/commissioning?beamline={beamline}"
    ).raise_for_status()
    commissioning_proposals = commissioning_proposals_response.json()[
        "commissioning_proposals"
    ]
    return proposal_number in commissioning_proposals


def validate_proposal(data_session_value, beamline) -> Dict[str, Any]:
    proposal_data = {}
    data_session_match = data_session_re.match(data_session_value)

    if data_session_match is None:
        raise ValueError(
            f"RE.md['data_session']='{data_session_value}' "
            f"is not matched by regular expression '{data_session_re.pattern}'"
        )

    try:
        current_cycle = get_current_cycle()
        proposal_number = data_session_match.group("proposal_number")
        proposal_commissioning = is_commissioning_proposal(proposal_number, beamline)
        proposal_response = nslsii_api_client.get(
            f"/v1/proposal/{proposal_number}"
        ).raise_for_status()
        proposal_data = proposal_response.json()["proposal"]
        if "error_message" in proposal_data:
            raise ValueError(
                f"while verifying data_session '{data_session_value}' "
                f"an error was returned by {proposal_response.url}: "
                f"{proposal_data}"
            )
        else:
            if (
                not proposal_commissioning
                and current_cycle not in proposal_data["cycles"]
            ):
                raise ValueError(
                    f"Proposal {data_session_value} is not valid in the current NSLS2 cycle ({current_cycle})."
                )
            if beamline.upper() not in proposal_data["instruments"]:
                raise ValueError(
                    f"Wrong beamline ({beamline.upper()}) for proposal {data_session_value} ({', '.join(proposal_data['instruments'])})."
                )
            # data_session is valid!

    except httpx.RequestError as rerr:
        # give the user a warning but allow the run to start
        warnings.warn(
            f"while verifying data_session '{data_session_value}' "
            f"the request {rerr.request.url!r} failed with "
            f"'{rerr}'"
        )

    return proposal_data


config_files = [
    os.path.expanduser("~/.config/n2sn_tools.yml"),
    "/etc/n2sn_tools.yml",
]


def authenticate(
    username,
):
    config = None
    for fn in config_files:
        try:
            with open(fn) as f:
                config = yaml.safe_load(f)
        except IOError:
            pass
        else:
            break

    if config is None:
        raise RuntimeError("Unable to open a config file")

    server = config.get("common", {}).get("server")

    if server is None:
        raise RuntimeError(f"Server name not found!")

    auth_server = Server(server, use_ssl=True)

    try:
        connection = Connection(
            auth_server,
            user=f"BNL\\{username}",
            password=getpass("Password : "),
            authentication=NTLM,
            auto_bind=True,
            raise_exceptions=True,
        )
        print(f"\nAuthenticated as : {connection.extend.standard.who_am_i()}")

    except LDAPInvalidCredentialsResult:
        raise RuntimeError(f"Invalid credentials for user '{username}'.") from None
    except LDAPSocketOpenError:
        print(f"{server} server connection failed...")


def should_they_be_here(username, new_data_session, beamline):
    user_access_json = nslsii_api_client.get(f"/v1/data-session/{username}").json()

    if "nsls2" in user_access_json["facility_all_access"]:
        return True

    elif beamline.lower() in user_access_json["beamline_all_access"]:
        return True

    elif new_data_session in user_access_json["data_sessions"]:
        return True

    return False


class AuthorizationError(Exception): ...


def sync_experiment(proposal_number, beamline, verbose=False, prefix=""):
    normalized_beamlines = {
        "sst1": "sst",
        "sst2": "sst",
    }
    redis_beamline = normalized_beamlines.get(beamline.lower(), beamline)
    redis_client = redis.Redis(host=f"info.{redis_beamline.lower()}.nsls2.bnl.gov")

    redis_prefix = f"{prefix}-" if prefix else ""
    md = RedisJSONDict(redis_client=redis_client, prefix=redis_prefix)

    new_data_session = f"pass-{proposal_number}"
    username = input("Username : ")

    if (new_data_session == md.get("data_session")) and (
        username == md.get("username")
    ):
        warnings.warn(
            f"Experiment {new_data_session} was already started by the same user."
        )

    else:
        proposal_data = validate_proposal(new_data_session, beamline)
        users = proposal_data.pop("users")

        authenticate(username)

        if not should_they_be_here(username, new_data_session, beamline):
            raise AuthorizationError(
                f"User '{username}' is not allowed to take data on proposal {new_data_session}"
            )

        pi_name = ""
        for user in users:
            if user.get("is_pi"):
                pi_name = (
                    f'{user.get("first_name", "")} {user.get("last_name", "")}'.strip()
                )

        md["data_session"] = new_data_session
        md["username"] = username
        md["start_datetime"] = datetime.now().isoformat()
        md["cycle"] = (
            "commissioning"
            if is_commissioning_proposal(str(proposal_number), beamline)
            else get_current_cycle()
        )
        md["proposal"] = {
            "proposal_id": proposal_data.get("proposal_id"),
            "title": proposal_data.get("title"),
            "type": proposal_data.get("type"),
            "pi_name": pi_name,
        }

        print(f"Started experiment {new_data_session}.")
        if verbose:
            print(json.dumps(proposal_data, indent=2))
            print("Users:")
            print(users)


def main():
    # Used by the `sync-experiment` command

    parser = argparse.ArgumentParser(
        description="Start or switch beamline experiment and record it in Redis"
    )
    parser.add_argument(
        "-b",
        "--beamline",
        dest="beamline",
        type=str,
        help="Which beamline (e.g. CHX)",
        required=True,
    )
    parser.add_argument(
        "-e",
        "--endstation",
        dest="prefix",
        type=str,
        help="Prefix for redis keys (e.g. by endstation)",
        required=False,
    )
    parser.add_argument(
        "-p",
        "--proposal",
        dest="proposal",
        type=int,
        help="Which proposal (e.g. 123456)",
        required=True,
    )
    parser.add_argument("-v", "--verbose", action=argparse.BooleanOptionalAction)

    args = parser.parse_args()
    sync_experiment(
        proposal_number=args.proposal,
        beamline=args.beamline,
        verbose=args.verbose,
        prefix=args.prefix,
    )



"""
Ophyd devices and signals for Delta Tau devices, such as Programmable Multi-Axis Controller (PMAC-PC)
"""

import logging

from ophyd import Component as Cpt
from ophyd import Device, EpicsSignal, EpicsSignalRO, Kind


class PMACStatus(EpicsSignalRO):
    """
    A signal to read the status of the PMAC-PC controller by checking the bit 11 of the status register.
    """

    def get(self):
        return int(super().get() >> 11 & 1 == 0)


class PMACKillSwitch(Device):
    """
    A device to kill the PMAC-PC controller.
    In the named version of the EPICS record (e.g., XF:23ID1-ES{Dif-Ax:Del}), the status is expected at Sts:4-Sts.
    Whereas in the axis numbered version (e.g., XF:23ID1-CT{MC:12-Ax:1}), the status is expected at Sts:1-Sts.
    """

    kill = Cpt(EpicsSignal, "Cmd:Kill-Cmd.PROC", kind=Kind.omitted)
    status = Cpt(PMACStatus, "Sts:4-Sts", kind=Kind.normal)

    def set(self, value, *args, **kwargs):
        """
        Set the kill switch to the given value
        """
        if value != 1:
            logging.getLogger(__name__).warning(
                "The value of the PMACKiller should only ever be set to 1. " "Changing the setpoint to 1 now."
            )
            value = 1
        self.kill.set(value, *args, **kwargs)

    def get(self, *args, **kwargs):
        """
        Get the value of the kill switch
        """
        return self.status.get(*args, **kwargs)



from datetime import date
from enum import Enum
from pathlib import Path
from typing import Optional
from ophyd_async.core import (
    FilenameProvider,
    PathProvider,
    PathInfo,
)
import os
import shortuuid


class YMDGranularity(int, Enum):
    none = 0
    year = 1
    month = 2
    day = 3


class ProposalNumYMDPathProvider(PathProvider):
    def __init__(
        self,
        filename_provider: FilenameProvider,
        metadata_dict: dict,
        granularity: YMDGranularity = YMDGranularity.day,
        separator=os.path.sep,
        **kwargs,
    ):
        self._filename_provider = filename_provider
        self._metadata_dict = metadata_dict
        self._granularity = granularity
        self._ymd_separator = separator
        self._beamline_proposals_dir = self.get_beamline_proposals_dir()
        super().__init__(filename_provider, **kwargs)

    @property
    def filename_provider(self):
        return self._filename_provider

    def get_beamline_proposals_dir(self):
        """
        Function that computes path to the proposals directory based on TLA env vars
        """

        beamline_tla = os.getenv(
            "ENDSTATION_ACRONYM", os.getenv("BEAMLINE_ACRONYM", "")
        ).lower()
        beamline_proposals_dir = Path(f"/nsls2/data/{beamline_tla}/proposals/")

        return beamline_proposals_dir

    def generate_directory_path(self, device_name: Optional[str] = None):
        """Helper function that generates ymd path structure"""

        current_date_template = ""
        if self._granularity == YMDGranularity.day:
            current_date_template = f"%Y{self._ymd_separator}%m{self._ymd_separator}%d"
        elif self._granularity == YMDGranularity.month:
            current_date_template = f"%Y{self._ymd_separator}%m"
        elif self._granularity == YMDGranularity.year:
            current_date_template = f"%Y{self._ymd_separator}"

        current_date = date.today().strftime(current_date_template)

        if device_name is None:
            ymd_dir_path = current_date
        else:
            ymd_dir_path = os.path.join(
                device_name,
                current_date,
            )

        directory_path = (
            self._beamline_proposals_dir
            / self._metadata_dict["cycle"]
            / self._metadata_dict["data_session"]
            / "assets"
            / ymd_dir_path
        )

        return directory_path

    def __call__(self, device_name: str = None) -> PathInfo:
        directory_path = self.generate_directory_path(device_name=device_name)

        return PathInfo(
            directory_path=directory_path,
            filename=self._filename_provider(),
            create_dir_depth=-self._granularity,
        )


class ProposalNumScanNumPathProvider(ProposalNumYMDPathProvider):
    def __init__(
        self,
        filename_provider: FilenameProvider,
        metadata_dict: dict,
        base_name: str = "scan",
        granularity: YMDGranularity = YMDGranularity.none,
        ymd_separator=os.path.sep,
        **kwargs,
    ):
        self._base_name = base_name
        super().__init__(
            filename_provider,
            metadata_dict,
            granularity=granularity,
            ymd_separator=ymd_separator,
            **kwargs,
        )

    def __call__(self, device_name: Optional[str] = None) -> PathInfo:
        directory_path = self.generate_directory_path(device_name=device_name)

        final_dir_path = (
            directory_path / f"{self._base_name}_{self._metadata_dict['scan_id']:06}"
        )

        return PathInfo(
            directory_path=final_dir_path,
            filename=self._filename_provider(),
            # Add one to dir depth creation level to account for scan dir
            create_dir_depth=-self._granularity - 1,
        )


class ShortUUIDFilenameProvider(FilenameProvider):
    """Generates short uuid filenames with device name as prefix"""

    def __init__(self, separator="_", **kwargs):
        self._separator = separator
        super().__init__(**kwargs)

    def __call__(self, device_name: Optional[str] = None) -> str:
        sid = shortuuid.uuid()
        if device_name is not None:
            return f"{device_name}{self._separator}{sid}"
        else:
            return sid


class AcqModeFilenameProvider(ShortUUIDFilenameProvider):
    def __init__(self, initial_mode, **kwargs):
        if not isinstance(initial_mode, Enum) or not isinstance(
            initial_mode.value, str
        ):
            raise TypeError("Initial acquisition mode type must be a string enum!")

        self._mode = initial_mode
        self._mode_type = type(initial_mode)
        super().__init__(**kwargs)

    def switch_mode(self, new_mode):
        if not isinstance(new_mode, self._mode_type):
            raise RuntimeError(
                f"{new_mode} is not a valid option for {self._mode_type}!"
            )
        else:
            self._mode = new_mode

    def __call__(self, **kwargs):
        return super().__call__(device_name=self._mode.value)


class DeviceNameFilenameProvider(FilenameProvider):
    """Filename provider that uses device name as filename"""

    def __call__(self, device_name: Optional[str] = None) -> str:
        if device_name is None:
            raise RuntimeError(
                f"Device name must be passed in when calling {type(self).__name__}!"
            )
        return device_name


class NSLS2PathProvider(ProposalNumYMDPathProvider):
    """
    Default NSLS2 path provider

    Generates paths in the following format:

    /nsls2/data/{TLA}/proposals/{CYCLE}/{PROPOSAL}/assets/{DETECTOR}/{Y}/{M}/{D}

    Filenames will be {DETECTOR}_{SHORT_UID} followed by the appropriate
    extension as determined by your detector writer.

    Parameters
    ----------
    metadata_dict : dict
        Typically `RE.md`. Used for dynamic save path generation from sync-d experiment
    """

    def __init__(self, *args, **kwargs):
        super().__init__(ShortUUIDFilenameProvider(), *args, **kwargs)
